use std::{
    fs::File,
    io::{Cursor, Read},
    path::{Path, PathBuf},
};

use tar::Archive;

use super::{Repository, arch::Architecture};

/// General Package
#[derive(Debug, Clone)]
pub struct Package {
    /// Repository of the package
    pub repo: String,
    /// `Architecture` of the package
    pub arch: Architecture,
    /// Name of the package
    pub name: String,
    pub rel: u64,
    /// Version of the package
    pub version: Option<String>,
}

impl Package {
    /// Create a new package
    pub fn new(repo: &str, arch: Architecture, pkg_name: &str, version: &str, rel: u64) -> Self {
        let pkg = Package {
            repo: repo.to_string(),
            arch,
            name: pkg_name.to_string(),
            rel: rel,
            version: Some(version.to_string()),
        };

        std::fs::create_dir_all(pkg.base_path()).unwrap();
        pkg
    }

    pub fn version(ver: &str) -> (String, u64) {
        let mut splitted = ver.split('-').collect::<Vec<_>>();
        let rel = splitted.pop().unwrap();
        let ver = splitted.join("-");
        (ver, rel.parse().unwrap())
    }

    pub fn install_script(&self) -> Option<String> {
        let pkg = self.base_path().join(self.file_name());
        read_file_tar(&pkg, ".INSTALL")
    }

    pub fn file_list(&self) -> Vec<String> {
        list_tar_file(&self.base_path().join(self.file_name()))
            .unwrap()
            .into_iter()
            .filter(|x| !x.ends_with("/"))
            .collect()
    }

    pub fn binaries(&self) -> Vec<String> {
        list_tar_file(&self.base_path().join(self.file_name()))
            .unwrap_or_default()
            .into_iter()
            .filter_map(|x| {
                let mut paths: Vec<_> = x.split("/").collect();
                paths.reverse();
                let parent = paths.get(1)?;
                if (*parent == "bin" || *parent == "sbin") && !x.ends_with("/") {
                    return Some(x);
                }

                None
            })
            .collect()
    }

    pub fn pkginfo_from_str(content: &str) -> Vec<(String, String)> {
        let mut ret: Vec<(String, Vec<String>)> = Vec::new();

        for line in content.split("\n") {
            if line.starts_with('#') || line.is_empty() {
                continue;
            }

            let (key, val) = line.split_once(" = ").unwrap();

            if let Some(e) = ret.iter_mut().find(|x| x.0 == key) {
                e.1.push(val.trim().to_string());
            } else {
                ret.push((key.to_string(), vec![val.trim().to_string()]));
            }
        }

        let mut ret: Vec<_> = ret.into_iter().map(|x| (x.0, x.1.join(";"))).collect();
        ret.sort_by(|a, b| a.0.cmp(&b.0));

        ret
    }

    pub fn pkginfo(&self) -> Vec<(String, String)> {
        let content = read_file_tar(&self.base_path().join(self.file_name()), ".PKGINFO").unwrap();
        Package::pkginfo_from_str(&content)
    }

    pub fn arch(&self) -> Vec<Architecture> {
        let mut ret = Vec::new();
        for a in [
            Architecture::x86_64,
            Architecture::aarch64,
            Architecture::any,
        ] {
            let check_pkg = self.switch_arch(a.clone());
            if check_pkg.exists() {
                ret.push(a);
            }
        }

        ret
    }

    /// Extract values from a package filename
    ///
    /// # Example
    ///
    /// ```
    /// use pacco::pkg::Package;
    ///
    /// let (name, version, rel, arch) = Package::extract_pkg_name("pkg-ver-rel-x86_64.pkg.tar.zst").unwrap();
    /// assert_eq!(name, "pkg");
    /// assert_eq!(version, "ver");
    /// assert_eq!(rel, "rel");
    /// assert_eq!(arch, pacco::pkg::arch::Architecture::x86_64);
    ///
    /// let (name, version, rel, arch) = Package::extract_pkg_name("my-pkg-ver-rel-x86_64.pkg.tar.zst").unwrap();
    /// assert_eq!(name, "my-pkg");
    /// assert_eq!(version, "ver");
    /// assert_eq!(rel, "rel");
    /// assert_eq!(arch, pacco::pkg::arch::Architecture::x86_64);
    /// ```
    pub fn extract_pkg_name(file_name: &str) -> Option<(String, String, String, Architecture)> {
        // Extract (assuming the filename is "<pkg_name>-<version>-<relation>-<arch>.pkg.tar.zst")
        let file_name = file_name.trim_end_matches(".sig").to_string();

        let mut splitted = file_name.split('-').collect::<Vec<&str>>();

        let arch = splitted.pop()?;
        assert!(arch.ends_with(".pkg.tar.zst"), "{file_name}");
        let arch = arch.trim_end_matches(".pkg.tar.zst");

        let relation = splitted.pop()?;
        let version = splitted.pop()?;

        let pkg_name = splitted.join("-");

        Some((
            pkg_name,
            version.to_string(),
            relation.to_string(),
            Architecture::parse(arch)?,
        ))
    }

    /// Parse a pkg filename
    pub fn from_filename(repo: &str, file_name: &str) -> Package {
        let (pkg_name, version, rel, arch) = Package::extract_pkg_name(file_name).unwrap();
        Self {
            repo: repo.to_string(),
            arch,
            name: pkg_name,
            rel: rel.parse().unwrap(),
            version: Some(version.to_string()),
        }
    }

    /// Find a package with latest version
    pub fn find(repo: &str, arch: Architecture, pkg_name: &str) -> Option<Self> {
        let mut base = Package {
            repo: repo.to_string(),
            arch,
            name: pkg_name.to_string(),
            rel: 1,
            version: None,
        };

        let versions = base.versions();
        let ver = versions.first()?;
        let (ver, rel) = Package::version(&ver);

        base.version = Some(ver.clone());
        base.rel = rel;

        Some(base)
    }

    pub fn man_entries(&self) -> Vec<String> {
        let files = self.file_list();
        files
            .into_iter()
            .filter(|x| x.starts_with("usr/share/man"))
            .map(|x| {
                x.trim_start_matches("usr/share/man/")
                    .trim_end_matches(".gz")
                    .to_string()
            })
            .collect()
    }

    pub fn kernel_modules(&self) -> Vec<String> {
        let files = self.file_list();
        files
            .into_iter()
            .filter(|x| x.starts_with("usr/lib/modules"))
            .map(|x| {
                x.trim_start_matches("usr/lib/modules/")
                    .trim_end_matches(".zst")
                    .to_string()
            })
            .collect()
    }

    pub fn fonts(&self) -> Vec<String> {
        let files = self.file_list();
        files
            .into_iter()
            .filter(|x| x.starts_with("usr/share/fonts"))
            .map(|x| x.trim_start_matches("usr/share/fonts/").to_string())
            .collect()
    }

    pub fn firmware(&self) -> Vec<String> {
        let files = self.file_list();
        files
            .into_iter()
            .filter(|x| x.starts_with("usr/lib/firmware"))
            .map(|x| {
                x.trim_start_matches("usr/lib/firmware/")
                    .trim_end_matches(".zst")
                    .to_string()
            })
            .collect()
    }

    pub fn keyrings(&self) -> Vec<String> {
        let files = self.file_list();
        files
            .into_iter()
            .filter(|x| x.starts_with("usr/share/pacman/keyrings"))
            .map(|x| {
                x.trim_start_matches("usr/share/pacman/keyrings/")
                    .to_string()
            })
            .collect()
    }

    pub fn shared_objects(&self) -> Vec<String> {
        list_tar_file(&self.base_path().join(self.file_name()))
            .unwrap_or_default()
            .into_iter()
            .filter(|x| {
                let file_name = x.split("/").last().unwrap();
                file_name.contains(".so.") || file_name.ends_with(".so")
            })
            .collect()
    }

    pub fn icons(&self) -> Vec<String> {
        let files = self.file_list();
        files
            .into_iter()
            .filter(|x| x.starts_with("usr/share/icons"))
            .map(|x| x.trim_start_matches("usr/share/icons/").to_string())
            .collect()
    }

    pub fn etc_entries(&self) -> Vec<String> {
        let files = self.file_list();
        files.into_iter().filter(|x| x.starts_with("etc")).collect()
    }

    pub fn systemd_units(&self) -> Vec<String> {
        // TODO : Extract unit infos
        list_tar_file(&self.base_path().join(self.file_name()))
            .unwrap_or_default()
            .into_iter()
            .filter(|x| {
                let ext = x.split(".").last().unwrap();
                ext == "service" || ext == "timer" || ext == "mount" || ext == "socket"
            })
            .collect()
    }

    pub fn readmes(&self) -> Vec<(String, String)> {
        let pkg_file = self.base_path().join(self.file_name());
        let files = self.file_list();
        files
            .into_iter()
            .filter(|x| {
                let file_name = x.split("/").last().unwrap();
                let cleaned = file_name.trim_end_matches(".md").trim_end_matches(".txt");
                cleaned == "README"
            })
            .map(|x| {
                let content = read_file_tar(&pkg_file, &x).unwrap();
                (x, content)
            })
            .collect()
    }

    pub fn pacman_hooks(&self) -> Vec<(String, String)> {
        let pkg_file = self.base_path().join(self.file_name());
        let files = self.file_list();
        files
            .into_iter()
            .filter(|x| {
                x.starts_with("etc/pacman.d/hooks/") || x.starts_with("usr/share/libalpm/hooks/")
            })
            .map(|x| {
                let content = read_file_tar(&pkg_file, &x).unwrap();
                (x, content)
            })
            .collect()
    }

    /// Save a new package to repository
    pub fn save(&self, pkg: Vec<u8>, sig: Option<Vec<u8>>) {
        let pkg_file = self.base_path().join(self.file_name());
        let sig_file = self.base_path().join(format!("{}.sig", self.file_name()));

        std::fs::write(&pkg_file, pkg).unwrap();
        if let Some(sig) = sig {
            std::fs::write(sig_file, sig).unwrap();
        }

        let db_file = PathBuf::from("./data")
            .join(&self.repo)
            .join(self.arch.to_string())
            .join(format!("{}.db.tar.gz", self.repo));

        repo_add(db_file.to_str().unwrap(), pkg_file.to_str().unwrap());

        // Add to any arch repo db if `arch=any`
        if self.arch == Architecture::any {
            let archs = Repository::new(&self.repo).unwrap().arch();

            for arch in archs {
                if arch == Architecture::any {
                    continue;
                }

                let db_file = PathBuf::from("./data")
                    .join(&self.repo)
                    .join(arch.to_string())
                    .join(format!("{}.db.tar.gz", self.repo));

                repo_add(db_file.to_str().unwrap(), pkg_file.to_str().unwrap());
            }
        }
    }

    fn base_path(&self) -> PathBuf {
        // <repo>/<arch>/<pkg>/
        Path::new("./data")
            .join(&self.repo)
            .join(self.arch.to_string())
            .join(&self.name)
    }

    /// Switch the `Architecture` of the package
    pub fn switch_arch(&self, arch: Architecture) -> Self {
        let mut new = self.clone();
        new.arch = arch;
        new
    }

    /// Switch to a specific `version` of the package
    pub fn get_version(&self, version: &str) -> Self {
        let mut new_pkg = self.clone();
        new_pkg.version = Some(version.to_string());
        new_pkg
    }

    /// Check if the specific package exists
    pub fn exists(&self) -> bool {
        let pkg_file = self.base_path().join(self.file_name());
        pkg_file.exists()
    }

    /// Checks if the package has a signature
    pub fn is_signed(&self) -> bool {
        let signed_file = self.base_path().join(format!("{}.sig", self.file_name()));
        signed_file.exists()
    }

    /// Build a pkg filename from the packages values
    pub fn file_name(&self) -> String {
        format!(
            "{}-{}-{}-{}.pkg.tar.zst",
            self.name,
            if let Some(ver) = &self.version {
                ver.to_string()
            } else {
                let versions = self.versions();
                Package::version(&versions.first().unwrap().clone()).0
            },
            self.rel,
            self.arch.to_string(),
        )
    }

    /// Get all versions of the package
    pub fn versions(&self) -> Vec<String> {
        let dir_path = self.base_path();
        let mut versions: Vec<String> = vec![];

        if let Ok(entries) = std::fs::read_dir(dir_path) {
            for entry in entries.filter_map(Result::ok) {
                let file_name = entry.file_name().into_string().unwrap_or_default();
                if file_name.starts_with(&self.name) && file_name.ends_with(".pkg.tar.zst") {
                    let (_, version, rel, _) = Package::extract_pkg_name(&file_name).unwrap();
                    versions.push(format!("{version}-{rel}"));
                }
            }
        }

        // Sort versions in descending order (most recent version first)
        versions.sort_by(|a, b| b.cmp(a));
        versions
    }

    /// Get the content of the `.pkg.tar.zst`
    pub fn pkg_content(&self) -> Option<Vec<u8>> {
        if self.exists() {
            return std::fs::read(self.base_path().join(self.file_name())).ok();
        }

        None
    }

    /// Get the content of the `.pkg.tar.zst.sig`
    pub fn sig_content(&self) -> Option<Vec<u8>> {
        if self.exists() {
            return std::fs::read(self.base_path().join(format!("{}.sig", &self.file_name()))).ok();
        }

        None
    }
}

pub fn run_command(cmd: Vec<&str>) {
    std::process::Command::new(cmd.first().unwrap())
        .args(cmd.into_iter().skip(1).collect::<Vec<&str>>())
        .spawn()
        .unwrap()
        .wait()
        .unwrap();
}

/// Add a package file to a repo db
pub fn repo_add(db_file: &str, pkg_file: &str) {
    run_command(vec!["repo-add", db_file, pkg_file]);
}

pub fn read_file_tar(tar: &PathBuf, file_path: &str) -> Option<String> {
    let mut file = File::open(tar).ok()?;
    let mut buf = Vec::new();
    file.read_to_end(&mut buf).unwrap();
    read_file_tar_raw(&buf, file_path)
}

pub fn read_file_tar_raw(tar: &[u8], file_path: &str) -> Option<String> {
    let uncompressed = zstd::decode_all(Cursor::new(tar)).unwrap();
    let content = Cursor::new(uncompressed);
    let mut a = Archive::new(content);

    for e in a.entries().ok()? {
        let mut e = e.ok()?;
        let path = e.path().unwrap();
        let path = path.to_str().unwrap();
        if path == file_path {
            let mut file_content = Vec::new();
            e.read_to_end(&mut file_content).unwrap();
            return String::from_utf8(file_content).ok();
        }
    }

    None
}

pub fn list_tar_file(tar: &PathBuf) -> Option<Vec<String>> {
    let mut file = File::open(tar).ok()?;
    let mut buf = Vec::new();
    file.read_to_end(&mut buf).unwrap();
    let uncompressed = zstd::decode_all(Cursor::new(buf)).unwrap();
    let content = Cursor::new(uncompressed);
    let mut a = Archive::new(content);
    let mut paths = Vec::new();

    for e in a.entries().ok()? {
        let e = e.ok()?;
        let path = e.path().ok()?;
        let path = path.to_str()?;
        paths.push(path.to_string());
    }

    Some(paths)
}
